#include <inttypes.h>
#include <stddef.h>
#include <stdint.h>

#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/adc.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/util.h>

#include "analog_in.h"
#include "retained.h"

#if !DT_NODE_EXISTS(DT_PATH(zephyr_user)) || \
	!DT_NODE_HAS_PROP(DT_PATH(zephyr_user), io_channels)
#error "No suitable devicetree overlay specified"
#endif

#define DT_SPEC_AND_COMMA(node_id, prop, idx) \
	ADC_DT_SPEC_GET_BY_IDX(node_id, idx),

/* Data of ADC io-channels specified in devicetree. */
static const struct adc_dt_spec adc_channels[] = {
	DT_FOREACH_PROP_ELEM(DT_PATH(zephyr_user), io_channels,
			     DT_SPEC_AND_COMMA)
};

uint16_t buf;
struct adc_sequence sequence = {
	.buffer = &buf,
	/* buffer size in bytes, not number of samples */
	.buffer_size = sizeof(buf),
};


int Analog_init(void)
{
	int err;
	for (size_t i = 0U; i < ARRAY_SIZE(adc_channels); i++) {
		if (!adc_is_ready_dt(&adc_channels[i])) {
			//printk("ADC controller device %s not ready\n", adc_channels[i].dev->name);
			return 0;
		}

		err = adc_channel_setup_dt(&adc_channels[i]);
		if (err < 0) {
			//printk("Could not setup channel #%d (%d)\n", i, err);
			return 0;
		}
	}
	return 0;
}

int16_t AnalogReadRaw(int channel)
{
	int16_t val_mv;
	int err;
	(void)adc_sequence_init_dt(&adc_channels[channel], &sequence);
	err = adc_read(adc_channels[channel].dev, &sequence);
	if (err < 0) {
		//printk("Could not read (%d)\n", err);
	}

	/*
	 * If using differential mode, the 16 bit value
	 * in the ADC sample buffer should be a signed 2's
	 * complement value.
	 */
	if (adc_channels[channel].channel_cfg.differential) {
		val_mv = ((int16_t)buf);
	} else {
		val_mv = (int16_t)buf;
	}
	return val_mv;
}

int32_t AnalogConvert_mV(int channel, int16_t raw)
{
	int32_t val_mv=raw;	
	int32_t vref_mv;
	uint8_t resolution;
	float val_float;
	const struct adc_dt_spec *spec=&adc_channels[channel];

	if (!spec->channel_cfg_dt_node_exists) {
		return -ENOTSUP;
	}

	if (spec->channel_cfg.reference == ADC_REF_INTERNAL) {
		vref_mv = (int32_t)adc_ref_internal(spec->dev);
	} else {
		vref_mv = spec->vref_mv;
	}

	resolution = spec->resolution;

	/*
	 * For differential channels, one bit less needs to be specified
	 * for resolution to achieve correct conversion.
	 */
	if (spec->channel_cfg.differential) {
		resolution -= 1U;
	}
	

	int32_t adc_mv = raw * vref_mv;
	//printk("res: %d, ref: %d, adc: %d\n", resolution, vref_mv, adc_mv);
	adc_gain_invert(spec->channel_cfg.gain, &adc_mv);
	//printk("res: %d, ref: %d, adc_gain: %d\n", resolution, vref_mv, adc_mv);

	switch(channel) {
		case ANALOG_INPUT_1:
			val_float = retained.cal_val[0] * ((adc_mv *((ANALOG_IN1_R1+ANALOG_IN1_R2)/ANALOG_IN1_R2)) );			
			break;
		case ANALOG_INPUT_DC_IN:
			val_float = retained.cal_val[1] * ((adc_mv *((ANALOG_DC_IN_R1+ANALOG_DC_IN_R2)/ANALOG_DC_IN_R2)) );
			break;
		case ANALOG_INPUT_BATTERY:
			val_float = retained.cal_val[2] * (adc_mv );
			break;
		case ANALOG_INPUT_20mA:
			val_float = retained.cal_val[3] * (adc_mv) ;
			break;
		default:
			//printk("Unconfigured channel!\n");
			return 0;
	}
	val_mv = (int32_t)val_float >> resolution;
	
	return val_mv;	
}